from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import ghidra
    from ghidra_builtins import *


def decomp_correlate(self, matches, p1_missing, p2_missing, p1_matches, p2_matches):
    """
    from all of the unmatched functions remaining, see if any should be matched by decomp
    This is slow, but sometimes necessary
    """

    # only attempt if there is something to match
    if len(p1_missing) > 0 and len(p2_missing) > 0:

        self.logger.info(f'Attempting to Decomp Correlate unmatched functions p1:{len(p1_missing)} p2:{len(p1_missing)}')

        for p1_func in p1_missing:

            # skip already matched functions
            if p1_matches.contains(p1_func.getEntryPoint()):
                continue

            decomp1 = self.enhance_sym(p1_func.getSymbol(), get_decomp_info=True)['code']

            for p2_func in p2_missing:

                # skip already matched functions
                if p2_matches.contains(p2_func.getEntryPoint()):
                    continue

                decomp2 = self.enhance_sym(p2_func.getSymbol(), get_decomp_info=True)['code']

                if self.remove_code_sig(decomp1) == self.remove_code_sig(decomp2):

                    # Correlate function Same Decomp
                    name = 'Decomp Match'
                    p1_addr = p1_func.getEntryPoint()
                    p2_addr = p2_func.getEntryPoint()
                    matches.setdefault((p1_addr, p2_addr), {}).setdefault(name, 0)
                    matches[(p1_addr, p2_addr)][name] += 1
                    p1_matches.add(p1_addr)
                    p2_matches.add(p2_addr)

                    # break to find the next match
                    break



from functools import lru_cache
import uuid
from typing import List, Tuple, TYPE_CHECKING

from jpype import JImplements, JOverride
from ghidra.app.plugin.match import FunctionHasher


if TYPE_CHECKING:
    import ghidra
    from ghidra_builtins import *

@JImplements(FunctionHasher, deferred=True)
class StructuralGraphHasher:
    """
    Hash function using Graph Centric Comparison of function control flow graphs (thank you 2004 halvar)
    Hash calculates a 3-tuple measurement from each node (function cfg) in the program
    (num_basic_blocks, num_edges_of_blocks, num_call_subfunctions)
    There are several other properties (length, refcount, paramcount) added to this hash, as it is meant to be run with one_to_many = True
    Based on "Structural Comparison of Executable Objects" by Halvar Flake
    """

    MIN_FUNC_LEN = 10
    MATCH_TYPE = 'StructuralGraphHash'

    @JOverride
    def hash(self, func: 'ghidra.program.model.listing.Function', monitor: 'ghidra.util.task.TaskMonitor') -> int:

        from ghidra.program.model.block import BasicBlockModel
        from ghidra.program.model.symbol import SourceType
        from ghidra.program.model.symbol import SymbolUtilities

        # graph structure vars
        num_basic_blocks = 0
        num_edges_of_blocks = 0
        num_call_subfunctions = 0

        sym = func.symbol

        # # skip func like FUN_ and LAB_, or local symbols like switch
        if sym.getSource() == SourceType.DEFAULT:
            fname = ''
        else:
            fname = SymbolUtilities.getCleanSymbolName(sym.getName(True), sym.address)

        basic_model = BasicBlockModel(func.getProgram(), True)
        basic_blocks = basic_model.getCodeBlocksContaining(func.getBody(), monitor)

        for block in basic_blocks:
            num_edges_of_blocks += block.getNumDestinations(monitor)
            num_basic_blocks += 1

            code_units = func.getProgram().getListing().getCodeUnits(block, True)
            for code in code_units:
                if code.mnemonicString == 'CALL' or code.mnemonicString == 'bl':
                    num_call_subfunctions += 1

        return hash((fname, num_basic_blocks, num_edges_of_blocks, num_call_subfunctions, func.body.numAddresses, func.parameterCount, sym.referenceCount))

    @ JOverride
    def commonBitCount(self, funcA: 'ghidra.program.model.listing.Function', funcB: 'ghidra.program.model.listing.Function', monitor: 'ghidra.util.task.TaskMonitor') -> int:
        raise NotImplementedError


@JImplements(FunctionHasher, deferred=True)
class StructuralGraphExactHasher:
    """
    Hash function using Graph Centric Comparison of function control flow graphs (thank you 2004 halvar)
    Hash calculates a 3-tuple measurement from each node (function cfg) in the program
    (num_basic_blocks, num_edges_of_blocks, num_call_subfunctions)
    Based on "Structural Comparison of Executable Objects" by Halvar Flake
    """

    MIN_FUNC_LEN = 10
    MATCH_TYPE = 'StructuralGraphExactHash'

    @JOverride
    def hash(self, func: 'ghidra.program.model.listing.Function', monitor: 'ghidra.util.task.TaskMonitor') -> int:

        from ghidra.program.model.block import BasicBlockModel

        num_basic_blocks = 0
        num_edges_of_blocks = 0
        num_call_subfunctions = 0

        basic_model = BasicBlockModel(func.getProgram(), True)
        basic_blocks = basic_model.getCodeBlocksContaining(func.getBody(), monitor)

        for block in basic_blocks:
            num_edges_of_blocks += block.getNumDestinations(monitor)
            num_basic_blocks += 1

            code_units = func.getProgram().getListing().getCodeUnits(block, True)
            for code in code_units:
                # TODO verify BL instruction for ARM https://developer.arm.com/documentation/den0013/d/Application-Binary-Interfaces/Procedure-Call-Standard?lang=en
                if code.mnemonicString == 'CALL' or code.mnemonicString == 'bl':
                    num_call_subfunctions += 1

        return hash((num_basic_blocks, num_call_subfunctions, num_edges_of_blocks))

    @ JOverride
    def commonBitCount(self, funcA: 'ghidra.program.model.listing.Function', funcB: 'ghidra.program.model.listing.Function', monitor: 'ghidra.util.task.TaskMonitor') -> int:
        raise NotImplementedError


@ JImplements(FunctionHasher, deferred=True)
class BulkInstructionsHasher:
    """
    Hash from instructions sorted and hashed
    Order of instructions is ignored
    Based on https://github.com/threatrack/ghidra-patchdiff-correlator#bulk-instructions-match
    """

    MIN_FUNC_LEN = 10
    MATCH_TYPE = 'BulkInstructionHash'

    @ JOverride
    def hash(self, func: 'ghidra.program.model.listing.Function', monitor: 'ghidra.util.task.TaskMonitor') -> int:

        instructions = []

        # reset iterator
        code_units = func.getProgram().getListing().getCodeUnits(func.getBody(), True)

        for code in code_units:
            instructions.append(code.toString())

        return hash(tuple(sorted(instructions)))

    @ JOverride
    def commonBitCount(self, funcA: 'ghidra.program.model.listing.Function', funcB: 'ghidra.program.model.listing.Function', monitor: 'ghidra.util.task.TaskMonitor') -> int:
        raise NotImplementedError


@ JImplements(FunctionHasher, deferred=True)
class BulkMnemonicHasher:
    """
    Hash from function mnemonics sorted
    Order of instructions is ignored
    Based on https://github.com/threatrack/ghidra-patchdiff-correlator#bulk-mnemonics-match
    """

    MIN_FUNC_LEN = 10
    MATCH_TYPE = 'BulkMnemonicHash'

    @ JOverride
    def hash(self, func: 'ghidra.program.model.listing.Function', monitor: 'ghidra.util.task.TaskMonitor') -> int:

        mnemonics = []

        # reset iterator
        code_units = func.getProgram().getListing().getCodeUnits(func.getBody(), True)

        for code in code_units:
            mnemonics.append(code.getMnemonicString())

        return hash(tuple(sorted(mnemonics)))

    @ JOverride
    def commonBitCount(self, funcA: 'ghidra.program.model.listing.Function', funcB: 'ghidra.program.model.listing.Function', monitor: 'ghidra.util.task.TaskMonitor') -> int:
        raise NotImplementedError


@ JImplements(FunctionHasher, deferred=True)
class BulkBasicBlockMnemonicHasher:
    """
    Hash from function basic blocks mnemonics sorted
    Order of basic blocks mnemonics is ignored
    Based on https://github.com/threatrack/ghidra-patchdiff-correlator#bulk-basic-block-mnemonics-match
    """

    MIN_FUNC_LEN = 10
    MATCH_TYPE = 'BulkBasicBlockMnemonicHash'

    @ JOverride
    def hash(self, func: 'ghidra.program.model.listing.Function', monitor: 'ghidra.util.task.TaskMonitor') -> int:

        from ghidra.program.model.block import BasicBlockModel

        basic_model = BasicBlockModel(func.getProgram(), True)
        basic_blocks = basic_model.getCodeBlocksContaining(func.getBody(), monitor)
        blocks = []

        for block in basic_blocks:
            code_units = func.getProgram().getListing().getCodeUnits(block, True)
            for code in code_units:
                blocks.append(code.getMnemonicString())

        return hash(tuple(sorted(blocks)))

    @ JOverride
    def commonBitCount(self, funcA: 'ghidra.program.model.listing.Function', funcB: 'ghidra.program.model.listing.Function', monitor: 'ghidra.util.task.TaskMonitor') -> int:
        raise NotImplementedError


@JImplements(FunctionHasher, deferred=True)
class NamespaceNameParamHasher:
    """
    Simply return Name with Namespace and Param Hash matches
    DO NOT RUN THIS with one_to_many = TRUE
    """

    MIN_FUNC_LEN = 10
    MATCH_TYPE = 'NamespaceNameParamHash'
    FIRST_RUN = True

    @JOverride
    def hash(self, func: 'ghidra.program.model.listing.Function', monitor: 'ghidra.util.task.TaskMonitor') -> int:

        return hash((func.getName(True), func.parameterCount))

    @ JOverride
    def commonBitCount(self, funcA: 'ghidra.program.model.listing.Function', funcB: 'ghidra.program.model.listing.Function', monitor: 'ghidra.util.task.TaskMonitor') -> int:
        raise NotImplementedError


@JImplements(FunctionHasher, deferred=True)
class NameParamHasher:
    """
    Simply return Name and Param Hash matches
    No namespace included
    DO NOT RUN THIS with one_to_many = TRUE
    """

    MIN_FUNC_LEN = 10
    MATCH_TYPE = 'NameParamHash'
    FIRST_RUN = True

    @JOverride
    def hash(self, func: 'ghidra.program.model.listing.Function', monitor: 'ghidra.util.task.TaskMonitor') -> int:

        return hash((func.getName(), func.parameterCount))

    @JOverride
    def commonBitCount(self, funcA: 'ghidra.program.model.listing.Function', funcB: 'ghidra.program.model.listing.Function', monitor: 'ghidra.util.task.TaskMonitor') -> int:
        raise NotImplementedError


@JImplements(FunctionHasher, deferred=True)
class NameParamRefHasher:
    """
    Hash based on name param and number of refs

    DO NOT RUN THIS with one_to_many = TRUE
    """

    MIN_FUNC_LEN = 10
    MATCH_TYPE = 'NameParamHash'
    FIRST_RUN = True

    @JOverride
    def hash(self, func: 'ghidra.program.model.listing.Function', monitor: 'ghidra.util.task.TaskMonitor') -> int:

        return hash((func.getName(True), func.parameterCount, func.symbol.referenceCount))

    @JOverride
    def commonBitCount(self, funcA: 'ghidra.program.model.listing.Function', funcB: 'ghidra.program.model.listing.Function', monitor: 'ghidra.util.task.TaskMonitor') -> int:
        raise NotImplementedError


@JImplements(FunctionHasher, deferred=True)
class SigCallingCalledHasher:
    """
    Hash based on signature, called, and calling functions ignoring FUN_*
    Not very reliable for programs lacking symbols

    DO NOT RUN THIS with one_to_many = TRUE
    This should be run very late in the game
    """

    MIN_FUNC_LEN = 10
    MATCH_TYPE = 'SigCallingCalledHasher'
    FIRST_RUN = True
    DEBUG = False

    @JOverride
    def hash(self, func: 'ghidra.program.model.listing.Function', monitor: 'ghidra.util.task.TaskMonitor') -> int:

        called = [called.toString() for called in func.getCalledFunctions(monitor) if "FUN_" not in called.toString()]
        calling = [calling.toString() for calling in func.getCalledFunctions(monitor)
                   if "FUN_" not in calling.toString()]

        sig = func.getSignature().toString().replace(func.name, '')

        if self.DEBUG:
            print(func.name)
            print((tuple(sorted(called)), tuple(sorted(calling)), sig))

        return hash((tuple(sorted(called)), tuple(sorted(calling)), sig))

    @JOverride
    def commonBitCount(self, funcA: 'ghidra.program.model.listing.Function', funcB: 'ghidra.program.model.listing.Function', monitor: 'ghidra.util.task.TaskMonitor') -> int:
        raise NotImplementedError


def getStringAtAddr(addr):
    """Get string at an address, if present"""
    from ghidra.program.model.data import StringDataType
    data = getDataAt(addr)
    if data is not None:
        dt = data.getDataType()
        if isinstance(dt, StringDataType):
            return str(data)
    return None


@lru_cache(None)
def get_defined_data(program: "ghidra.program.model.listing.Program"):

    from ghidra.program.model.symbol import SymbolUtilities
    from ghidra.program.model.symbol import SourceType
    from ghidra.program.model.symbol import SymbolType
    from ghidra.program.model.listing import Function

    def _is_sym_string(sym: 'ghidra.program.model.symbol.Symbol') -> bool:
        is_string = False
        sym_addr = sym.getAddress()
        if sym_addr is not None:
            data = sym.getProgram().getListing().getDataAt(sym_addr)
            if data is not None and data.hasStringValue():
                is_string = True
        return is_string

    strings_in_func = []
    func_str_map = {}

    # from java.util import CollectionUtils
    from ghidra.program.model.data import StringDataInstance
    from ghidra.program.util import DefinedDataIterator

    for data in DefinedDataIterator.definedStrings(program):
        sdi_str = StringDataInstance.getStringDataInstance(data)
        s = sdi_str.getStringValue()
        if s != None:
            strings_in_func.append(str(s))

    for sym in program.getSymbolTable().getAllSymbols(True):
        if not sym.referenceCount == 0:
            # skip functions
            if not sym.symbolType == SymbolType.FUNCTION:  # and _is_sym_string(sym):
                # don't include DEFAULT (FUN_ LAB_) but do include strings s_something
                # if sym.getSource() != SourceType.DEFAULT:
                sym_addr = sym.getAddress()
                if sym_addr is not None:
                    data = sym.getProgram().getListing().getDataAt(sym_addr)
                    if data is not None and data.hasStringValue():
                        # its a string, find which functions use it
                        for ref in sym.references:
                            # print(ref.referenceType.toString())
                            f = program.getFunctionManager().getFunctionContaining(ref.fromAddress)
                            if f is not None:
                                func_str_map.setdefault(f.entryPoint, []).append(str(data))

    return func_str_map


@JImplements(FunctionHasher, deferred=True)
class StringsRefsHasher:
    """
    Hash based on signature, called, and calling functions ignoring FUN_*

    DO NOT RUN THIS with one_to_many = TRUE
    This should be run very late in the game
    """

    MIN_FUNC_LEN = 10
    MIN_STRING_LEN = 5
    MATCH_TYPE = 'StringsRefsHasher'
    ONE_TO_MANY = True  # supports one to many
    DEBUG = False

    @JOverride
    def hash(self, func: 'ghidra.program.model.listing.Function', monitor: 'ghidra.util.task.TaskMonitor') -> int:

        func_str_map = get_defined_data(func.getProgram())

        # print(get_defined_data.cache_info())

        strings = func_str_map.get(func.entryPoint)

        if strings is not None:
            strings = sorted(strings)
        else:
            # assign unique value for functions without strings. This allows
            strings = [uuid.uuid4()]

        if self.DEBUG:
            print(strings)
            print(len(strings))

        return hash((tuple(strings)))

    @JOverride
    def commonBitCount(self, funcA: 'ghidra.program.model.listing.Function', funcB: 'ghidra.program.model.listing.Function', monitor: 'ghidra.util.task.TaskMonitor') -> int:
        raise NotImplementedError


@JImplements(FunctionHasher, deferred=True)
class StrUniqueFuncRefsHasher:
    """
    Hash based on signature, called, and calling functions ignoring FUN_*

    DO NOT RUN THIS with one_to_many = TRUE
    This should be run very late in the game
    """

    MIN_FUNC_LEN = 10
    MIN_STRING_LEN = 5
    MATCH_TYPE = 'StrUniqueFuncRefsHasher'
    ONE_TO_MANY = False  # supports one to many
    DEBUG = False

    @JOverride
    def hash(self, func: 'ghidra.program.model.listing.Function', monitor: 'ghidra.util.task.TaskMonitor') -> int:

        func_str_map = get_defined_data(func.getProgram())

        # print(get_defined_data.cache_info())

        strings = func_str_map.get(func.entryPoint)
        ref_count = func.getSymbol().getReferenceCount()

        if strings is not None:
            strings = sorted(list(set(strings)))

        else:
            # assign unique value for functions without strings. This allows
            strings = [uuid.uuid4()]

        if self.DEBUG:
            print(strings)
            print(len(strings))
            print(ref_count)

        return hash((tuple(strings), ref_count))

    @JOverride
    def commonBitCount(self, funcA: 'ghidra.program.model.listing.Function', funcB: 'ghidra.program.model.listing.Function', monitor: 'ghidra.util.task.TaskMonitor') -> int:
        raise NotImplementedError


@lru_cache(None)
def get_func_to_switch(program: "ghidra.program.model.listing.Program"):

    func_switch_map = {}

    for sym in program.getSymbolTable().getAllSymbols(True):
        sym: 'ghidra.program.model.symbol.Symbol' = sym

        if sym.getName().startswith('switchD') or sym.getName().startswith('caseD') or "switchD" in sym.getName():
            f = program.getFunctionManager().getFunctionContaining(sym.address)
            # if f is None:
            #     for ref in sym.references:
            #         # print(ref.referenceType.toString())
            #         f = program.getFunctionManager().getFunctionContaining(ref.fromAddress)
            #         if f is not None:
            #             func_switch_map.setdefault(f.entryPoint, []).append(str(sym))
            # else:
            #     func_switch_map.setdefault(f.entryPoint, []).append(str(sym))
            if f is not None:
                func_switch_map.setdefault(f.entryPoint, []).append(str(sym))
            for ref in sym.references:
                # print(ref.referenceType.toString())
                f = program.getFunctionManager().getFunctionContaining(ref.fromAddress)
                if f is not None:
                    func_switch_map.setdefault(f.entryPoint, []).append(str(sym))

    return func_switch_map


@JImplements(FunctionHasher, deferred=True)
class SwitchSigHasher:
    """
    Simply return Name and Param Hash matches
    No namespace included
    DO NOT RUN THIS with one_to_many = TRUE
    """

    MIN_FUNC_LEN = 10
    MATCH_TYPE = 'SwitchSigHasher'
    ONE_TO_MANY = False
    DEBUG = False

    @JOverride
    def hash(self, func: 'ghidra.program.model.listing.Function', monitor: 'ghidra.util.task.TaskMonitor') -> int:

        sig = func.getSignature().toString().replace(func.name, '')

        func_switch_map = get_func_to_switch(func.getProgram())

        switch_syms = func_switch_map.get(func.entryPoint)

        if switch_syms is not None:
            switch_syms = sorted(switch_syms)
        else:
            # assign unique value for functions without strings.
            switch_syms = [uuid.uuid4()]

        if self.DEBUG:
            print(func.name)
            print(switch_syms)

        return hash(tuple([sig, tuple(switch_syms)]))

    @ JOverride
    def commonBitCount(self, funcA: 'ghidra.program.model.listing.Function', funcB: 'ghidra.program.model.listing.Function', monitor: 'ghidra.util.task.TaskMonitor') -> int:
        raise NotImplementedError



import difflib
import re
from textwrap import dedent
from typing import List, Tuple, Union, TYPE_CHECKING
import logging
import json

from mdutils.tools.Table import Table
from mdutils.mdutils import MdUtils
from mdutils.tools.TableOfContents import TableOfContents


class GhidriffMarkdown:

    def __init__(self, logging=logging.INFO) -> None:
        self.logger = self.setup_logger(level=logging)

    def setup_logger(self, level: int = logging.INFO) -> logging.Logger:
        """
        Setup Class Instance Logger
        """
        logging.basicConfig(
            format='%(levelname)-5s| %(name)s | %(message)s',
            datefmt='%H:%M:%S'
        )

        logger = logging.getLogger(__package__)
        logger.setLevel(level)

        return logger

    def _wrap_with_diff(self, diff: str) -> str:

        text = ''
        text += "```diff\n"
        text += diff
        text += "\n```\n"
        text += "\n"

        return text

    def _wrap_with_code(self, code: str, style='') -> str:
        text = ''
        text += f"```{style}" + "\n"
        text += code
        text += "\n```\n"
        text += "\n"

        return text

    def _wrap_with_details(self, diff: str, summary: str = None) -> str:

        text = ''
        text += "<details>\n"
        if summary:
            text += f"<summary>{summary}</summary>\n"
        text += '\n'
        text += diff
        text += "\n</details>\n"

        return text

    def _wrap_list_with_details(self, items: list, max_items: int = 10) -> str:

        data = None
        if isinstance(items, list):
            if len(items) > max_items:
                show_items = [item for i, item in enumerate(items) if i <= max_items]
                hide_itmes = [item for i, item in enumerate(items) if i > max_items]
                data = f'<details><summary>Expand for full list:<br>{"<br>".join(show_items)}</summary>{"<br>".join(hide_itmes)}</details>'
            else:
                data = f'{"<br>".join(items)}'
        else:
            # do nothing
            data = items

        return data

    def gen_esym_table(self, old_name, esym, max_items=10) -> str:

        table_list = []
        table_list.extend(['Key', old_name])
        column_len = len(table_list)

        skip_keys = ['code', 'instructions', 'mnemonics', 'blocks', 'parent']
        count = 1
        for key in esym:
            if key in skip_keys:
                continue

            if isinstance(esym[key], list):
                data = self._wrap_list_with_details(esym[key], max_items)
            else:
                data = esym[key]

            table_list.extend([key, data])
            count += 1

        diff_table = Table().create_table(columns=column_len, rows=count, text=table_list, text_align='center')

        return diff_table

    def gen_esym_table_diff(self, old_name, new_name, modified, max_items=10) -> str:
        diff_table = ''

        table_list = []
        table_list.extend(['Key', old_name, new_name])
        column_len = len(table_list)

        skip_keys = ['code', 'instructions', 'mnemonics', 'blocks', 'parent']
        count = 1
        for key in modified['old']:
            if key in skip_keys:
                continue
            if key in modified['diff_type']:
                diff_key = f"`{key}`"
            else:
                diff_key = f"{key}"

            if isinstance(modified['old'][key], list):
                data = self._wrap_list_with_details(modified['old'][key], max_items)
                data2 = self._wrap_list_with_details(modified['new'][key], max_items)
            else:
                data = modified['old'][key]
                data2 = modified['new'][key]

            table_list.extend([diff_key, data, data2])

            # table_list.extend([diff_key, modified['old'][key], modified['new'][key]])
            count += 1

        diff_table = Table().create_table(columns=column_len, rows=count,
                                          text=table_list, text_align='center')

        return diff_table

    def gen_esym_table_diff_meta(self, old_name, new_name, modified) -> str:
        diff_table = ''

        table_list = []
        table_list.extend(['Key', f"{old_name} - {new_name}"])
        column_len = len(table_list)

        keys = ['diff_type', 'ratio', 'i_ratio', 'm_ratio', 'b_ratio', 'match_types']
        count = 1
        for key in keys:
            val = modified[key]
            if isinstance(val, list):
                val = ','.join(val)
            table_list.extend([key, val])
            count += 1

        diff_table = Table().create_table(columns=column_len, rows=count, text=table_list, text_align='center')

        return diff_table

    def gen_esym_key_diff(self, esym: dict, esym2: dict, key: str, exclude=None, n=3) -> str:
        """
        Generate a difflib unified diff from two esyms and a key
        n is the number of context lines for diff lib to wrap around the found diff
        """
        diff = ''

        if exclude:
            items = [item for item in esym[key] if re.search(exclude, item) is None]
            items2 = [item for item in esym2[key] if re.search(exclude, item) is None]
        else:
            items = esym[key]
            items2 = esym2[key]

        diff += '\n'.join(difflib.unified_diff(items, items2,
                          fromfile=f"{esym['fullname']} {key}", tofile=f"{esym2['fullname']} {key}", lineterm='', n=n))

        return self._wrap_with_diff(diff)

    @staticmethod
    def gen_combined_sxs_html_from_pdiff(pdiff: dict) -> str:
        """
        pdiff: Standard pdiff from ghidriff
        """

        def _add_header(line: str):
            # print(line)
            pass

        sxs_diff_htmls = []

        for mod in pdiff['functions']['modified']:

            if 'code' not in mod['diff_type']:
                continue

            old_code = mod['old']['code']
            new_code = mod['new']['code']
            old_name = mod['old']['fullname']
            new_name = mod['new']['fullname']

            sxs_diff_html = GhidriffMarkdown.gen_code_table_diff_html(
                old_code, new_code, old_name, new_name)

            sxs_diff_htmls.append([mod['old']['name'], sxs_diff_html])

        table_htmls = [table[1] for table in sxs_diff_htmls]

        charset = 'utf-8'

        html_diff = difflib.HtmlDiff(tabsize=4)
        html = (html_diff._file_template % dict(
            styles=html_diff._styles,
            legend=dedent(html_diff._legend),
            table="\n".join(table_htmls),
            charset=charset)).encode(charset, 'xmlcharrefreplace').decode(charset)

        for line in html.splitlines(True):

            _add_header(line)

        return html

    @staticmethod
    def gen_sxs_html_from_pdiff(pdiff: dict) -> list:
        """
        pdiff: Standard pdiff from ghidriff
        """

        sxs_diff_htmls = []

        for mod in pdiff['functions']['modified']:

            if 'code' not in mod['diff_type']:
                continue

            old_code = mod['old']['code']
            new_code = mod['new']['code']
            old_name = mod['old']['fullname']
            new_name = mod['new']['fullname']

            sxs_diff_html = GhidriffMarkdown.gen_code_table_diff_html(
                old_code, new_code, old_name, new_name, bottom=pdiff['html_credits'])

            sxs_diff_htmls.append([mod['old']['name'], sxs_diff_html])

        return sxs_diff_htmls

    @ staticmethod
    def gen_code_table_diff_html(old_code,
                                 new_code,
                                 old_name,
                                 new_name,
                                 html_type='file',
                                 dedent_table: bool = True,
                                 bottom=None,
                                 tabsize=2
                                 ) -> str:
        """
        Generates side by side diff in HTML
        dedent_table: True will dedent the indented table so it renders the html table in markdown
        type: 
            - inline Simply return table,style,and legend withiout html and body tags
            - table only - just return table with no style
            - full - return difflib html template complete
        """

        charset = 'utf-8'

        if isinstance(old_code, str):
            old_code = old_code.splitlines(True)
        if isinstance(new_code, str):
            new_code = new_code.splitlines(True)

        if html_type == 'inline':
            styles = '<style type="text/css">%(styles)s\n</style>' % dict(
                styles=difflib.HtmlDiff(tabsize=tabsize)._styles)
            table = difflib.HtmlDiff(tabsize=tabsize).make_table(
                old_code, new_code, fromdesc=old_name, todesc=new_name)

            diff_html = styles + table
            diff_html.encode(charset, 'xmlcharrefreplace').decode(charset)
        elif html_type == 'table-only':
            diff_html = difflib.HtmlDiff(tabsize=4).make_table(
                old_code, new_code, fromdesc=old_name, todesc=new_name)
        elif html_type == 'file':
            diff_html = difflib.HtmlDiff(tabsize=tabsize).make_file(
                old_code, new_code, fromdesc=old_name, todesc=new_name)
        else:
            raise NotImplementedError

        if dedent_table:
            diff_html = diff_html.splitlines(True)

            # handle dedent of table so that it renders in markdown
            start_table = None
            end_table = None
            for i, line in enumerate(diff_html):
                if line.find('<table class="diff"') != -1:
                    start_table = i
                if line.find('</table>') != -1:
                    end_table = i

                if start_table and end_table:
                    break

            pre_table = ''.join(diff_html[:start_table])
            post_table = ''.join(diff_html[end_table+1:])

            table = dedent(''.join(diff_html[start_table:end_table+1]))

            diff_html = pre_table + table + post_table

        if html_type == 'file' and bottom:
            needle = '</body>'
            loc = diff_html.find(needle)
            diff_html = diff_html[:loc] + bottom + diff_html[loc:]

        return diff_html + '\n'

    def gen_table_from_dict(self, headers: list, items: dict):

        table = ''

        table_list = []
        table_list.extend(headers)
        column_len = len(table_list)

        count = 1
        for key, values in items.items():
            table_list.extend([key, values])
            count += 1

        table = Table().create_table(columns=column_len, rows=count, text=table_list, text_align='center')

        return table

    def gen_strings_ref_table(self, str_data) -> str:

        table_list = []
        table_list.extend(['String', 'Ref Count', 'Ref Func'])
        column_len = len(table_list)

        count = 1
        for item in str_data:
            
            name = item['name']
            ref_count = item['refcount']
            calling = ','.join(item['calling'])
            table_list.extend([name,ref_count,calling])
            count += 1

        diff_table = Table().create_table(columns=column_len, rows=count, text=table_list, text_align='center')

        return diff_table
    
    def gen_strings_diff(self, deleted_strings: dict, added_strings: dict):

        def _clean_string(name: str):
            clean = re.sub(r'_[_a-f0-9]+$','',name)
            return clean.rstrip('_')

        added = sorted([f'{_clean_string(item["name"])}\n' for item in added_strings])

        deleted = sorted([f'{_clean_string(item["name"])}\n' for item in deleted_strings])

        diff = ''.join(list(difflib.unified_diff(deleted, added,
                                                 lineterm='\n', fromfile='deleted strings', tofile='added strings')))

        return self._wrap_with_diff(diff)

    def gen_metadata_diff(
        self,
        pdiff: Union[str, dict]
    ) -> str:
        """Generate binary metadata diff"""

        if isinstance(pdiff, str):
            pdiff = json.loads(pdiff)

        old_meta = pdiff['old_meta']
        new_meta = pdiff['new_meta']

        old_text = ''
        old_name = f"{old_meta['Program Name']} Meta"

        new_text = ''
        new_name = f"{new_meta['Program Name']} Meta"

        for i in old_meta:
            self.logger.debug(f"{i}: {old_meta[i]}")
            old_text += f"{i}: {old_meta[i]}\n"

        for i in new_meta:
            self.logger.debug(f"{i}: {new_meta[i]}")
            new_text += f"{i}: {new_meta[i]}\n"

        diff = ''.join(list(difflib.unified_diff(old_text.splitlines(True), new_text.splitlines(
            True), lineterm='\n', fromfile=old_name, tofile=new_name, n=100)))

        return diff

    def gen_mermaid_diff_flowchart(self, pdiff: dict, max_section_funcs: int = 25) -> str:

        diff_flow = '''
```mermaid

flowchart LR

{modified_links}

subgraph {new_bin}
    {new_modified}
    {added_sub}
end

subgraph {old_bin}
    {old_modified}
    {deleted_sub}
end

```'''

        added = []
        deleted = []
        modified_links = []
        old_modified = []
        new_modified = []

        old_bin = pdiff['old_meta']['Program Name']
        new_bin = pdiff['new_meta']['Program Name']

        for i, func in enumerate(pdiff['functions']['added']):
            name = func['fullname'].replace('::', '-')
            added.append(self._clean_md_header(name))

            if max_section_funcs and i > max_section_funcs:
                msg = f"{len(pdiff['functions']['added']) - max_section_funcs}_more_added_funcs_omitted..."
                added.append(self._clean_md_header(msg))
                break

        for i, func in enumerate(pdiff['functions']['deleted']):
            name = func['fullname'].replace('::', '-')
            deleted.append(self._clean_md_header(name))

            if max_section_funcs and i > max_section_funcs:
                msg = f"{len(pdiff['functions']['deleted']) - max_section_funcs}_more_deleted_funcs_omitted..."
                deleted.append(self._clean_md_header(msg))
                break

        modified_code_funcs = [func for func in pdiff['functions']['modified'] if 'code' in func['diff_type']]

        for i, modified in enumerate(modified_code_funcs):

            if max_section_funcs and i > max_section_funcs:
                modified_links.append(
                    f"{old_bin}<--{len(modified_code_funcs) - max_section_funcs}ommited-->{new_bin}")
                break

            old_modified.append(self._clean_md_header(
                f"{modified['old']['fullname']}-{modified['old']['paramcount']}-old"))
            new_modified.append(self._clean_md_header(
                f"{modified['new']['fullname']}-{modified['old']['paramcount']}-new"))
            modified_links.append(
                f"{self._clean_md_header(modified['old']['fullname'])}-{modified['old']['paramcount']}-old<--Match {int(modified['b_ratio']*100)}%-->{self._clean_md_header(modified['new']['fullname'])}-{modified['old']['paramcount']}-new")

        deleted_sub = ''
        added_sub = ''
        if len(deleted) > 0:
            deleted_sub = '''subgraph Deleted\ndirection LR\n{}\nend'''.format('\n    '.join(deleted))
        if len(added) > 0:
            added_sub = '''subgraph Added\ndirection LR\n{}\nend'''.format('\n    '.join(added))

        return diff_flow.format(old_bin=old_bin, new_bin=new_bin, added_sub=added_sub, deleted_sub=deleted_sub, modified_links='\n'.join(modified_links), old_modified='\n'.join(old_modified), new_modified='\n'.join(new_modified))

    def gen_mermaid_pie_from_dict(self, data: dict, title: str, skip_keys: list = None, include_keys: list = None) -> str:
        """
        Generate basic mermaidjs Pie chart from dict
        skip_keys: [ 'skipkey1', 'skipkey45'] List of keys to skip from Dict
        includes_keys: ['random_key1', 'otherkey2'] - Only include these keys
        Default: include all keys and values from dict.
        """

        pie_template = '''
```mermaid
pie showData
    title {title}
{rows}
```
'''
        rows = []

        for key, value in data.items():

            row = None

            if skip_keys and key in skip_keys:
                continue

            if include_keys:
                if key in include_keys:
                    row = f'"{self._clean_md_header(key)}" : {value}'
            else:
                row = f'"{self._clean_md_header(key)}" : {value}'

            if row:
                rows.append(row)

        return pie_template.format(title=title, rows='\n'.join(rows))

    def _clean_md_header_lower(self, text):
        return re.sub(r'[^a-z0-9_\\-]', '', text.lower().replace(' ', '-'))

    def _clean_md_header(self, text):
        return re.sub(r'[^A-Za-z0-9_\\-]', '', text.replace(' ', '-'))

    def gen_pe_download_cmd(self, pdiff: dict) -> str:
        """
        Generates Windows PE file download command        
        """

        def _decode_arch(proc, addr_size):
            arch = None
            if proc == 'x86':
                if addr_size == '64':
                    arch = 'x64'
                else:
                    arch = proc
            else:
                if addr_size == '64':
                    arch = 'arm64'
                else:
                    arch = 'arm'

            return arch

        old_url = pdiff['old_pe_url']
        new_url = pdiff['new_pe_url']

        # PE Property[OriginalFilename]: localspl.dll

        if pdiff['old_meta'].get('OriginalFilename') is not None:
            # handle Ghidra pe_key before 10.2
            pe_key = 'OriginalFilename'
            ver_key = 'ProductVersion'
        else:
            pe_key = 'PE Property[OriginalFilename]'
            ver_key = 'PE Property[ProductVersion]'

        old_filename = pdiff['old_meta'][pe_key].lower()
        new_filename = pdiff['new_meta'][pe_key].lower()

        # PE Property[ProductVersion]: 10.0.22000.795
        old_ver = pdiff['old_meta'][ver_key]
        new_ver = pdiff['new_meta'][ver_key]

        # Processor: x86
        # Address Size: 64
        old_arch = _decode_arch(pdiff['old_meta']['Processor'], pdiff['old_meta']['Address Size'])
        new_arch = _decode_arch(pdiff['new_meta']['Processor'], pdiff['new_meta']['Address Size'])

        old_dl = f"wget {old_url} -O {old_filename}.{old_arch}.{old_ver}"
        new_dl = f"wget {new_url} -O {new_filename}.{new_arch}.{new_ver}"

        dl_cmd = "\n".join((old_dl, new_dl))

        return dl_cmd

    def gen_diff_md(
        self,
        pdiff: Union[str, dict],
        title=None,
        side_by_side: bool = False,
        max_section_funcs: int = None,
        include_code: bool = False
    ) -> str:
        """
        Generate Markdown Diff from pdiff match results
        """

        self.logger.info(f"Generating markdown from {pdiff['stats']}")

        # use max passed into function, revert to class if it exists
        if not max_section_funcs:
            max_section_funcs = self.max_section_funcs

        self.logger.debug(f"max_section_funcs: {max_section_funcs}")

        if isinstance(pdiff, str):
            pdiff = json.loads(pdiff)

        funcs = pdiff['functions']

        old_name = pdiff['old_meta']['Program Name']
        new_name = pdiff['new_meta']['Program Name']

        if title:
            title = title
        else:
            title = f"{old_name}-{new_name} Diff"

        md = MdUtils('diff', title=title)

        # change title to atx style
        md.title = md.header.choose_header(level=1, title=title, style='atx')

        md.new_header(1, 'Visual Chart Diff')
        md.new_paragraph(self.gen_mermaid_diff_flowchart(pdiff))
        md.new_paragraph(self.gen_mermaid_pie_from_dict(
            pdiff['stats'], f"Function Matches - {pdiff['stats']['func_match_overall_percent']}",
            include_keys=['matched_funcs_len', 'unmatched_funcs_len']))
        md.new_paragraph(self.gen_mermaid_pie_from_dict(
            pdiff['stats'],
            f"Matched Function Similarity - {pdiff['stats']['match_func_similarity_percent']}",
            include_keys=['matched_funcs_with_code_changes_len', 'matched_funcs_with_non_code_changes_len', 'matched_funcs_no_changes_len']))

        # Create Metadata section
        md.new_header(1, 'Metadata')

        md.new_header(2, 'Ghidra Diff Engine')

        md.new_header(3, 'Command Line')
        known_cmd, extra_cmd, full_cmd = self.gen_diff_cmd_line(old_name, new_name)
        md.new_header(4, 'Captured Command Line', add_table_of_contents='n')
        md.new_paragraph(self._wrap_with_code(known_cmd))
        md.new_header(4, 'Verbose Args', add_table_of_contents='n')
        md.new_paragraph(self._wrap_with_details(self._wrap_with_code(full_cmd)))

        if pdiff.get('old_pe_url') is not None and pdiff.get('new_pe_url') is not None:
            md.new_header(4, 'Download Original PEs')
            md.new_paragraph(self._wrap_with_code(self.gen_pe_download_cmd(pdiff)))

        md.new_header(2, 'Binary Metadata Diff')
        md.new_paragraph(self._wrap_with_diff(self.gen_metadata_diff(pdiff)))

        md.new_header(2, 'Program Options')
        skip_options = 'Program Information'  # options duplicate metadta
        for key in pdiff['program_options'].keys():
            for opt_name in pdiff['program_options'][key]:
                if opt_name in skip_options:
                    continue
                # md.new_header(3, f'Ghidra {key} {opt_name.capitalize()} Options', add_table_of_contents='n')
                if pdiff['program_options'][key][opt_name] is not None:
                    md.new_paragraph(self._wrap_with_details(self.gen_table_from_dict(
                        [f'{opt_name.capitalize()} Option', 'Value'], pdiff['program_options'][key][opt_name]), f'Ghidra {key} {opt_name.capitalize()} Options'))

                else:
                    md.new_paragraph(f'*No {opt_name.capitalize()} set.*')

        md.new_header(2, 'Diff Stats')
        md.new_paragraph(self.gen_table_from_dict(['Stat', 'Value'], pdiff['stats']))
        md.new_paragraph(self.gen_mermaid_pie_from_dict(pdiff['stats']['match_types'], 'All Matches'))
        md.new_paragraph(self.gen_mermaid_pie_from_dict(pdiff['stats']['first_matches'], 'First Matches'))
        md.new_paragraph(self.gen_mermaid_pie_from_dict(pdiff['stats'], 'Diff Stats', include_keys=[
                         'added_funcs_len', 'deleted_funcs_len', 'modified_funcs_len']))
        md.new_paragraph(self.gen_mermaid_pie_from_dict(
            pdiff['stats'], 'Symbols', include_keys=['added_symbols_len', 'deleted_symbols_len']))

        # Create Strings Section
        md.new_header(2, 'Strings')
        if len(pdiff['strings']['deleted']) > 0 or len(pdiff['strings']['added']):
            md.new_paragraph(self.gen_mermaid_pie_from_dict(
                pdiff['stats'], 'Strings', include_keys=['added_strings_len', 'deleted_strings_len']))

            md.new_header(3, 'Strings Diff', add_table_of_contents='n')
            md.new_paragraph(self.gen_strings_diff(pdiff['strings']['deleted'], pdiff['strings']['added']))
            md.new_header(3, 'String References', add_table_of_contents='n')
            md.new_header(4, 'Old', add_table_of_contents='n')
            md.new_paragraph(self.gen_strings_ref_table(pdiff['strings']['deleted']))
            md.new_header(4, 'New', add_table_of_contents='n')
            md.new_paragraph(self.gen_strings_ref_table(pdiff['strings']['added']))
            
        else:
            md.new_paragraph('*No string differences found*\n')

        # Create Deleted section
        md.new_header(1, 'Deleted')

        for i, esym in enumerate(funcs['deleted']):

            if i > max_section_funcs:
                md.new_header(2, 'Max Deleted Section Functions Reached Error')
                md.new_line(f"{len(funcs['deleted']) - max_section_funcs} Deleted Functions Ommited...")
                self.logger.warn(f'Max Deleted Section Functions {max_section_funcs} Reached')
                self.logger.warn(f"{len(funcs['deleted']) - max_section_funcs} Functions Ommited...")
                break

            md.new_header(2, esym['fullname'])
            md.new_header(3, "Function Meta", add_table_of_contents='n')
            md.new_paragraph(self.gen_esym_table(old_name, esym))

            # only show code if not above section max
            if len(funcs['deleted']) < max_section_funcs:
                old_code = esym['code'].splitlines(True)
                new_code = ''.splitlines(True)
                diff = ''.join(list(difflib.unified_diff(old_code, new_code,
                                                         lineterm='\n', fromfile=esym['fullname'], tofile=esym['fullname'])))
                if len(diff) > 0:
                    md.new_paragraph(self._wrap_with_diff(diff))
                else:
                    md.new_paragraph(f"*No code available for {esym['fullname']}*")

        # Create Added section
        md.new_header(1, 'Added')

        for i, esym in enumerate(funcs['added']):

            if i > max_section_funcs:
                md.new_header(2, 'Max Added Section Functions Reached Error')
                md.new_line(f"{len(funcs['added']) - max_section_funcs} Added Functions Ommited...")
                self.logger.warn(f'Max Added Section Functions {max_section_funcs} Reached')
                self.logger.warn(f"{len(funcs['added']) - max_section_funcs} Functions Ommited...")
                break

            md.new_header(2, esym['fullname'])
            md.new_header(3, "Function Meta", add_table_of_contents='n')
            md.new_paragraph(self.gen_esym_table(new_name, esym))

            # only show code if not above section max
            if len(funcs['added']) < max_section_funcs:
                old_code = ''.splitlines(True)
                new_code = esym['code'].splitlines(True)
                diff = ''.join(list(difflib.unified_diff(old_code, new_code,
                                                         lineterm='\n', fromfile=esym['fullname'], tofile=esym['fullname'])))
                if len(diff) > 0:
                    md.new_paragraph(self._wrap_with_diff(diff))
                else:
                    md.new_paragraph(f"*No code available for {esym['fullname']}*")

        # Create Modified section
        md.new_header(1, 'Modified')
        md.new_paragraph(f"*Modified functions contain code changes*")
        modified_with_code = 0
        for modified in funcs['modified']:

            if modified_with_code > max_section_funcs:
                md.new_header(2, 'Max Modified Section Functions Reached Error')
                md.new_line(f"{len(funcs['modified']) - max_section_funcs} Functions Ommited...")
                self.logger.warn(f'Max Modified Section Functions {max_section_funcs} Reached')
                self.logger.warn(f"{len(funcs['modified']) - max_section_funcs} Functions Ommited...")
                break

            diff = None
            old_func_name = modified['old']['fullname']

            # selectively include matches
            if 'code' in modified['diff_type']:

                modified_with_code += 1

                md.new_header(2, old_func_name)

                md.new_header(3, "Match Info", add_table_of_contents='n')
                md.new_paragraph(self.gen_esym_table_diff_meta(old_name, new_name, modified))

                md.new_header(3, "Function Meta Diff", add_table_of_contents='n')
                md.new_paragraph(self.gen_esym_table_diff(old_name, new_name, modified))

                if 'called' in modified['diff_type']:
                    md.new_header(3, f"{old_func_name} Called Diff", add_table_of_contents='n')
                    md.new_paragraph(self.gen_esym_key_diff(modified['old'], modified['new'], 'called', n=0))
                if 'calling' in modified['diff_type']:
                    md.new_header(3, f"{old_func_name} Calling Diff", add_table_of_contents='n')
                    md.new_paragraph(self.gen_esym_key_diff(modified['old'], modified['new'], 'calling', n=0))

                md.new_header(3, f"{old_func_name} Diff", add_table_of_contents='n')
                md.new_paragraph(self._wrap_with_diff(modified['diff']))

                if include_code:
                    md.new_header(3, f"{old_func_name} Code", add_table_of_contents='n')
                    md.new_paragraph(self._wrap_with_code(modified['old']['code'], 'c'))
                    md.new_paragraph(self._wrap_with_code(modified['new']['code'], 'c'))

                # only include side by side diff if requested (this adds html to markdown and considerable size)
                if side_by_side:
                    md.new_header(3, f"{old_func_name} Side By Side Diff", add_table_of_contents='n')
                    html_diff = GhidriffMarkdown.gen_code_table_diff_html(
                        modified['old']['code'], modified['new']['code'], old_name, new_name, html_type='inline')
                    md.new_paragraph(html_diff)

        # Create Slightly Modified secion
        # slightly as in no code changes but other relevant changes.
        slight_mods = ['refcount', 'length', 'called', 'calling', 'name', 'fullname']

        md.new_header(1, 'Modified (No Code Changes)')
        md.new_paragraph(f"*Slightly modified functions have no code changes, rather differnces in:*")
        md.new_list(slight_mods)

        # skip this section (as it is mostly a bonus) if this markdown is already too big
        too_big = max_section_funcs < (len(funcs['added']) + len(funcs['deleted']) + modified_with_code)

        if too_big:
            md.new_header(2, 'Section Skipped')
            md.new_paragraph(
                f"**This section was skipped because markdown was too big. Adjust max_section_funcs: {max_section_funcs} to a higher number.**")
        else:
            for modified in funcs['modified']:

                if modified_with_code > max_section_funcs:
                    md.new_header(2, 'Max Modified (No Code Changes) Section Functions Reached Error')
                    md.new_line(f"{len(funcs['modified']) - max_section_funcs} Functions Ommited...")
                    self.logger.warn(f'Max Modified Section Functions {max_section_funcs} Reached')
                    self.logger.warn(f"{len(funcs['modified']) - max_section_funcs} Functions Ommited...")
                    break

                mods = set(slight_mods).intersection(set(modified['diff_type']))

                # only show non code mods and skip if code m_ratio is same
                if 'code' not in modified['diff_type'] and len(mods) > 0 and (modified['m_ratio'] < 1.0 or len(modified['diff']) == 0):

                    modified_with_code +=1
                    
                    if modified['old']['external']:
                        old_func_name = modified['old']['fullname']
                        new_func_name = modified['old']['fullname']
                    else:
                        old_func_name = modified['old']['name']
                        new_func_name = modified['old']['name']

                    if old_func_name.startswith('FUN_') or new_func_name.startswith('FUN_'):

                        ignore_called = False
                        ignore_calling = False

                        if len(modified['old']['called']) > 0 and len(modified['new']['called']) > 0:
                            called_set = set(modified['old']['called']).difference(modified['new']['called'])
                            ignore_called = all('FUN_' in name for name in list(called_set))

                        if len(modified['old']['calling']) > 0 and len(modified['new']['calling']) > 0:
                            calling_set = set(modified['old']['calling']).difference(modified['new']['calling'])
                            ignore_calling = all('FUN_' in name for name in list(calling_set))

                        # skip name and fullname changes
                        if len(mods.difference(['name', 'fullname'])) == 0:
                            continue
                        # if all called are FUN_ skip
                        elif 'called' in modified['diff_type'] and 'calling' in modified['diff_type'] and ignore_called and ignore_calling:
                            continue
                        elif 'calling' in modified['diff_type'] and ignore_calling:
                            continue
                        elif 'called' in modified['diff_type'] and called_set:
                            continue

                    md.new_header(2, old_func_name)

                    md.new_header(3, "Match Info", add_table_of_contents='n')
                    md.new_paragraph(self.gen_esym_table_diff_meta(old_name, new_name, modified))

                    md.new_header(3, "Function Meta Diff", add_table_of_contents='n')
                    md.new_paragraph(self.gen_esym_table_diff(old_name, new_name, modified))

                    if 'called' in modified['diff_type']:
                        md.new_header(3, f"{old_func_name} Called Diff", add_table_of_contents='n')
                        md.new_paragraph(self.gen_esym_key_diff(modified['old'], modified['new'], 'called', n=0))
                    if 'calling' in modified['diff_type']:
                        md.new_header(3, f"{old_func_name} Calling Diff", add_table_of_contents='n')
                        md.new_paragraph(self.gen_esym_key_diff(modified['old'], modified['new'], 'calling', n=0))

        # add credit
        md.new_paragraph(pdiff['md_credits'])

        # generate TOC and set style to atx
        md.table_of_contents += md.header.choose_header(level=1, title='TOC', style='atx')
        md.table_of_contents += TableOfContents().create_table_of_contents(md._table_titles, depth=3)

        return md.get_md_text().strip()



import pytest
from pathlib import Path

from ghidriff import GhidraDiffEngine, VersionTrackingDiff, get_parser

SYMBOLS_DIR = 'symbols'

@pytest.mark.forked
def test_parsing_pe_garbage(shared_datadir: Path):

    parser = get_parser()

    args = parser.parse_args(['test', 'test2']) # these args will not be tested

    
    not_a_pe_path = shared_datadir / SYMBOLS_DIR / 'pingme.txt'

    DiffEngine: GhidraDiffEngine = VersionTrackingDiff
    d = DiffEngine(args=args)

    assert None == d.get_pe_download_url(not_a_pe_path,'junk')


from argparse import ArgumentParser
from pathlib import Path
import json

from ghidriff import GhidraDiffEngine
from .parser import get_parser, get_engine_classes


def main():
    """
    ghidriff - GhidraDiffEngine module main function
    """

    parser: ArgumentParser = get_parser()

    GhidraDiffEngine.add_ghidra_args_to_parser(parser)

    args = parser.parse_args()

    output_path = Path(args.output_path)
    output_path.mkdir(exist_ok=True, parents=True)

    if args.log_path == 'None':
        engine_log_path = None
    if args.log_path == parser.get_default('log_path'):
        engine_log_path = output_path / parser.get_default('log_path')
    else:
        engine_log_path = Path(args.log_path)

    if args.project_location == parser.get_default('project_location'):
        project_path = output_path / parser.get_default('project_location')
    else:
        project_path = Path(args.project_location)

    if args.symbols_path == parser.get_default('symbols_path'):
        symbols_path = output_path / parser.get_default('symbols_path')
    else:
        symbols_path = Path(args.symbols_path)

    binary_paths = args.old + [bin for sublist in args.new for bin in sublist]

    binary_paths = [Path(path) for path in binary_paths]

    if any([not path.exists() for path in binary_paths]):
        missing_bins = [f'{path.name}' for path in binary_paths if not path.exists()]
        raise FileNotFoundError(f"Missing Bins: {' '.join(missing_bins)}")

    project_name = f'{args.project_name}-{binary_paths[0].name}-{binary_paths[-1].name}'

    engines = get_engine_classes()
    DiffEngine: GhidraDiffEngine = engines[args.engine]

    d: GhidraDiffEngine = DiffEngine(args=args,
                                     verbose=True,
                                     threaded=args.threaded,
                                     max_ram_percent=args.max_ram_percent,
                                     print_jvm_flags=args.print_flags,
                                     jvm_args=args.jvm_args,
                                     force_analysis=args.force_analysis,
                                     force_diff=args.force_diff,
                                     verbose_analysis=args.va,
                                     no_symbols=args.no_symbols,
                                     engine_log_path=engine_log_path,
                                     engine_log_level=args.log_level,
                                     engine_file_log_level=args.file_log_level,
                                     min_func_len=args.min_func_len,
                                     use_calling_counts=args.use_calling_counts,
                                     bsim=args.bsim,
                                     bsim_full=args.bsim_full,
                                     gdts=args.gdt
                                     )

    d.setup_project(binary_paths, project_path, project_name, symbols_path)

    d.analyze_project()

    diffs = []

    # pair up binaries with the n-1 version
    for i in range(len(binary_paths)-1):
        diffs.append((binary_paths[i], binary_paths[i+1]))

    # add a diff of the first and last binary for full coverage
    if not binary_paths[1] == binary_paths[-1] and args.summary:
        diffs.append((binary_paths[0], binary_paths[-1]))

    for diff in diffs:
        pdiff = d.diff_bins(diff[0], diff[1])
        pdiff_json = json.dumps(pdiff)

        d.validate_diff_json(pdiff_json)

        diff_name = f"{Path(diff[0]).name}-{Path(diff[1]).name}.ghidriff"

        d.dump_pdiff_to_path(diff_name,
                             pdiff,
                             output_path,
                             side_by_side=args.side_by_side,
                             max_section_funcs=args.max_section_funcs,
                             md_title=args.md_title)


if __name__ == "__main__":
    main()



__version__ = '0.7.2'
__author__ = 'clearbluejar'

# Expose API
from .ghidra_diff_engine import GhidraDiffEngine
from .version_tracking_diff import VersionTrackingDiff
from .simple_diff import SimpleDiff
from .structural_graph_diff import StructualGraphDiff
from .parser import get_parser, get_engine_classes

__all__ = [
    "GhidraDiffEngine", "SimpleDiff", "StructualGraphDiff", "VersionTrackingDiff", "get_parser", "get_engine_classes"
]



import hashlib
from pathlib import Path
from struct import unpack
from functools import lru_cache


@lru_cache(None)
def sha1_file(path: str) -> str:
    # https://stackoverflow.com/questions/22058048/hashing-a-file-in-python
    # BUF_SIZE is totally arbitrary, change for your app!
    BUF_SIZE = 65536  # lets read stuff in 64kb chunks!

    path = Path(path)

    sha1 = hashlib.sha1()

    with path.open('rb') as f:
        while True:
            data = f.read(BUF_SIZE)
            if not data:
                break
            sha1.update(data)

    return f'{sha1.hexdigest()}'


# get pe data for url gen
# https://github.com/m417z/winbindex/blob/gh-pages/data/extract_data_from_pe_files.py
def get_pe_extra_data(path: Path):

    path = Path(path)

    with path.open('rb') as handle:
        # Get PE offset (@60, DWORD) from DOS header
        handle.seek(60, 0)
        offset = handle.read(4)
        offset = unpack('<I', offset)[0]

        handle.seek(offset + 4, 0)
        word = handle.read(2)
        machine = unpack('<H', word)[0]

        handle.seek(offset + 8, 0)
        dword = handle.read(4)
        timestamp = unpack('<I', dword)[0]

        handle.seek(offset + 0x50, 0)
        dword = handle.read(4)
        image_size = unpack('<I', dword)[0]

    return {
        'machine': machine,
        'timestamp': timestamp,
        'image_size': image_size,
    }


def get_microsoft_download_url(filename, timestamp, virtual_size):

    assert filename is not None
    assert timestamp is not None
    assert virtual_size is not None

    timestamp = format(timestamp, '08X')
    virtual_size = format(virtual_size, 'X')

    return f'https://msdl.microsoft.com/download/symbols/{filename}/{timestamp}{virtual_size}/{filename}'


# utils from Pyhidra

def get_private_class(path: str):
    from java.lang import ClassLoader
    from jpype import JClass

    gcl = ClassLoader.getSystemClassLoader()
    return JClass(path, loader=gcl)


def set_field(cls, fname, value, obj=None):
    cls = cls.class_
    field = cls.getDeclaredField(fname)
    field.setAccessible(True)
    field.set(obj, value)
    field.setAccessible(False)



import inspect
import argparse
import ghidriff


def get_engine_classes() -> dict:
    engines = {}

    for name, klass in inspect.getmembers(ghidriff, inspect.isclass):
        if name.endswith('Diff'):
            engines[name] = klass

    return engines

def get_parser() -> argparse.ArgumentParser:
    """
    Build main ghidriff parser
    """

    parser = argparse.ArgumentParser(description='ghidriff - A Command Line Ghidra Binary Diffing Engine',
                                formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('old', nargs=1, help="Path to old version of binary '/somewhere/bin.old'")
    parser.add_argument('new', action='append', nargs='+',
                        help="Path to new version of binary '/somewhere/bin.new'. (For multiple new binaries add oldest to newest)")

    # setup Engine class options
    engines = get_engine_classes()
    parser.add_argument('--engine', help='The diff implementation to use.',
                        default='VersionTrackingDiff', choices=engines.keys())

    parser.add_argument('-o', '--output-path', help='Output path for resulting diffs', default='ghidriffs')
    parser.add_argument('--summary', help='Add a summary diff if more than two bins are provided', default=False)

    return parser


from pathlib import Path
import json
import pytest

from ghidriff import get_parser, get_engine_classes, VersionTrackingDiff, GhidraDiffEngine

SYMBOLS_DIR = 'symbols'
BINS_DIR = 'bins'


@pytest.mark.forked
def test_diff_ntoskrnl_decomp_unmatched(shared_datadir: Path):
    """
    Tests end to end diff of CVE
    runs forked because each jpype jvm can only be initialized 1x
    """

    test_name = 'ntoskrnl_decomp_unmatched'
    output_path = shared_datadir / test_name
    output_path.mkdir(exist_ok=True, parents=True)
    symbols_path = shared_datadir / SYMBOLS_DIR
    bins_path = shared_datadir / BINS_DIR
    ghidra_project_path = output_path / 'ghidra_projects'
    ghidra_project_path.mkdir(exist_ok=True,parents=True)

    # setup bins

    old_bin_path = bins_path / 'ntoskrnl.exe.x64.10.0.22621.2792.10-1-5.gzf'
    new_bin_path = bins_path / 'ntoskrnl.exe.x64.10.0.22621.2861.10-1-5.gzf'

    assert old_bin_path.exists()
    assert new_bin_path.exists()

    parser = get_parser()

    GhidraDiffEngine.add_ghidra_args_to_parser(parser)

    args = parser.parse_args(['-s', str(symbols_path), str(old_bin_path.absolute()), str(new_bin_path.absolute()), '-p', str(ghidra_project_path.absolute())])

    engine_log_path = output_path / parser.get_default('log_path')

    binary_paths = args.old + [bin for sublist in args.new for bin in sublist]

    binary_paths = [Path(path) for path in binary_paths]

    if any([not path.exists() for path in binary_paths]):
        missing_bins = [f'{path.name}' for path in binary_paths if not path.exists()]
        raise FileNotFoundError(f"Missing Bins: {' '.join(missing_bins)}")

    project_name = f'{args.project_name}-{binary_paths[0].name}-{binary_paths[-1].name}'

    DiffEngine: GhidraDiffEngine = VersionTrackingDiff

    d: GhidraDiffEngine = DiffEngine(args=args,
                                     verbose=True,
                                     threaded=args.threaded,
                                     max_ram_percent=args.max_ram_percent,
                                     print_jvm_flags=args.print_flags,
                                     jvm_args=args.jvm_args,
                                     force_analysis=args.force_analysis,
                                     force_diff=args.force_diff,
                                     verbose_analysis=args.va,
                                     no_symbols=args.no_symbols,
                                     engine_log_path=engine_log_path,
                                     engine_log_level=args.log_level,
                                     engine_file_log_level=args.file_log_level,
                                     )

    d.setup_project(binary_paths, args.project_location, project_name, args.symbols_path)

    d.analyze_project()

    pdiff = d.diff_bins(old_bin_path, new_bin_path)
    pdiff_json = json.dumps(pdiff)

    d.validate_diff_json(pdiff_json)

    diff_name = f"{old_bin_path.name}-{new_bin_path.name}_diff"

    d.dump_pdiff_to_path(diff_name,
                         pdiff,
                         output_path,
                         side_by_side=args.side_by_side,
                         max_section_funcs=args.max_section_funcs,
                         md_title=args.md_title)


    assert pdiff['stats']['match_types']['Decomp Match'] == 29

    assert len(pdiff['functions']['added']) == 0
    assert len(pdiff['functions']['deleted']) == 0
    assert len(pdiff['functions']['modified']) in [33, 34, 35]  # Various modified across Ghidra vers



# Ghidra/Features/VersionTrackingBSim/src/main/java/ghidra/feature/vt/api/BSimProgramCorrelator.java

# this should be called after several reliable correlators have taken place
# we need reliable matches to seed BSIM
# auto tracking lets you run autoversiontracking first (which runs all 8 correlators)
# then it runs BSIM

from .utils import get_private_class, set_field


def correlate_bsim(matches, p1, p2, p1_matches, p2_matches, monitor, logger=None, seed_match_types=None, p1_addr_set=None, p2_addr_set=None, enabled=True):
    """
    matches: previously matched functions used for seeds
    seed_match_types: the types of matches to consider for seeds
    """

    if not enabled:
        logger.info(f"Skipping BSIM correlator. BSIM disabled with arg --no-bsim")
        return

    from ghidra.feature.vt import api as vtapi

    if not hasattr(vtapi, 'BSimProgramCorrelatorFactory'):
        logger.info(f"Skipping BSIM correlator. BSIM not present in this version of Ghidra")
        return

    from ghidra.feature.vt.api.main import VTMatchInfo, VTAssociationType, VTSession, VTScore
    from ghidra.feature.vt.api import BSimProgramCorrelatorFactory, BSimProgramCorrelator
    from ghidra.feature.vt.api.db import VTSessionDB
    from ghidra.util import SystemUtilities

    # see Ghidra/Features/VersionTracking/src/main/java/ghidra/feature/vt/gui/task/CreateManualMatchTask.java#L62

    def _create_match_info(match, match_set, p1, p2) -> VTMatchInfo:
        info: VTMatchInfo = VTMatchInfo(match_set)

        func1 = p1.getFunctionManager().getFunctionAt(match[0])
        func2 = p2.getFunctionManager().getFunctionAt(match[1])

        info.setSourceAddress(func1.getEntryPoint())
        info.setDestinationAddress(func2.getEntryPoint())
        info.setSourceLength(int(func1.getBody().getNumAddresses()))
        info.setDestinationLength(int(func2.getBody().getNumAddresses()))
        info.setSimilarityScore(VTScore(1.0))
        info.setConfidenceScore(VTScore(1.0))
        info.setAssociationType(VTAssociationType.FUNCTION)

        return info

    logger.info("Starting BSIM correlator")

    # create a VT session and match set for bsim "accepted matches" seed

    from java.lang import Object

    bsim_factory = BSimProgramCorrelatorFactory()
    options = bsim_factory.createDefaultOptions()
    session_ref = Object()

    # Start fix Ghidra 11.1 for ghidriff #95
    reset_mode = False

    if not SystemUtilities.isInTestingMode():
        su = get_private_class('ghidra.util.SystemUtilities')
        set_field(su, "isInTestingMode", True)
        reset_mode = True

    session: VTSession = VTSessionDB.createVTSession(bsim_factory.name, p1, p2, session_ref)

    if reset_mode:
        set_field(su, "isInTestingMode", False)
        assert SystemUtilities.isInTestingMode() == False
    # end fix for #95

    # create match set of already accepted matches using fake correlator
    transaction = session.startTransaction('seed')
    # match_set = session.createMatchSet(symbol_correlator)
    match_set = session.getManualMatchSet()

    if seed_match_types is None:
        seed_match_types = ['SymbolsHash', 'ExactBytesFunctionHasher', 'ExactInstructionsFunctionHasher',
                            'StructuralGraphExactHasher', 'ExactMnemonicsFunctionHasher']

    for match, m_types in matches.items():
        # if the match type is in the allowed seed_types create a seed match
        if any(m_type in seed_match_types for m_type in m_types):
            vt_match_info = _create_match_info(match, match_set, p1, p2)
            match_set.addMatch(vt_match_info)

    # Print all match sets from session
    match_sets = session.getMatchSets()
    for match_set in match_sets:
        logger.info(f'{match_set}')

    # BSIM will seed using accepted matches Ghidra/Features/VersionTrackingBSim/src/main/java/ghidra/feature/vt/api/BSimProgramCorrelatorMatching.java#L558-L595
    for match in match_set.getMatches():
        match.association.setAccepted()

    session.endTransaction(int(transaction), True)

    # instantiate bsim and find matches

    transaction = session.startTransaction(bsim_factory.name)

    # if not AddrSetView is defined, use the entire loaded mem
    if p1_addr_set is None:
        p1_addr_set = p1.memory.loadedAndInitializedAddressSet
    if p2_addr_set is None:
        p2_addr_set = p2.memory.loadedAndInitializedAddressSet

    bsim_correlator: BSimProgramCorrelator = bsim_factory.createCorrelator(p1, p1_addr_set, p2, p2_addr_set, options)
    bsim_correlator.correlate(session, monitor)
    session.endTransaction(int(transaction), True)

    # Print all match sets from session
    match_sets = session.getMatchSets()
    for match_set in match_sets:

        # updated ghidriff matches with BSIM findings
        if match_set.getProgramCorrelatorName() == bsim_factory.name:

            logger.info(f'{match_set}')

            for bsim_match in match_set.getMatches():
                # Correlate function as Implied Match
                name = 'BSIM'
                matches.setdefault((bsim_match.sourceAddress, bsim_match.destinationAddress), {}).setdefault(name, 0)
                matches[(bsim_match.sourceAddress, bsim_match.destinationAddress)][name] += 1
                p1_matches.add(bsim_match.sourceAddress)
                p2_matches.add(bsim_match.destinationAddress)

    # release object?
    session.release(session_ref)



import hashlib

from typing import List, Tuple, TYPE_CHECKING

from .ghidra_diff_engine import GhidraDiffEngine

if TYPE_CHECKING:
    import ghidra
    from ghidra_builtins import *


class StructualGraphDiff(GhidraDiffEngine):
    """
    An Ghidra Diff implementation using simple comparison mechanisms
    """

    def find_matches(
        self,
        p1: "ghidra.program.model.listing.Program",
        p2: "ghidra.program.model.listing.Program",
        ignore_FUN: bool = False,
    ) -> dict:
        """
        Find matching and unmatched functions between p1 and p2
        """

        from ghidra.program.util import DiffUtility

        compare_key_memo = {}

        def _get_compare_key(sym: 'ghidra.program.model.symbol.Symbol', func: 'ghidra.program.model.listing.Function') -> tuple:
            """
            Builds tuple from symbol (parent, name, refcount, length, paramcount)
            """
            key = f'{sym.getID()}-{sym.getProgram().getName()}'

            if compare_key_memo.get(key) is None:

                fhash = ''
                cu_count = 0
                if (not func.isThunk() and func.getBody().getNumAddresses() >= 10):
                    # reset iterator
                    code_units = func.getProgram().getListing().getCodeUnits(func.getBody(), True)

                    mnemonics = []

                    # Mnemonic Bulker
                    for code in code_units:
                        mnemonics.append(code.getMnemonicString())
                        cu_count += 1

                    fhash = hashlib.sha256(''.join(mnemonics).encode('UTF-8')).hexdigest()

                compare_key_memo[key] = (cu_count, sym.getReferenceCount(),
                                         func.body.numAddresses, func.parameterCount, fhash)

            return compare_key_memo[key]

        def _get_compare_key2(sym: 'ghidra.program.model.symbol.Symbol', func: 'ghidra.program.model.listing.Function') -> tuple:
            """
            Builds structural graph hash. Thank you Halvar Flake!)
            """

            from ghidra.program.model.block import BasicBlockModel
            from ghidra.util.task import ConsoleTaskMonitor

            # graph structure vars
            num_basic_blocks = 0
            num_edges_of_blocks = 0
            num_call_subfunctions = 0

            if (not func.isThunk() and func.getBody().getNumAddresses() >= 10):

                monitor = ConsoleTaskMonitor()

                basic_model = BasicBlockModel(func.getProgram(), True)
                basic_blocks = basic_model.getCodeBlocksContaining(func.getBody(), monitor)

                for block in basic_blocks:
                    num_edges_of_blocks += block.getNumDestinations(monitor)
                    num_basic_blocks += 1

                    code_units = func.getProgram().getListing().getCodeUnits(block, True)
                    for code in code_units:
                        if code.getMnemonicString() == 'CALL':
                            num_call_subfunctions += 1

            # ignore Ghidra generated function names
            gen_sym_prefix = ['FUN_', 'lambda', 'destructor', '~', 'dynamic_initializer', 'dtor', '::']
            fname = sym.getName(True)

            if any([generated in fname for generated in gen_sym_prefix]):
                fname = ''
            # ignore sections of names with relative addresses or offsets
            # Catch_All@64616295 _GrowCompBitsBuffer@4
            elif '@' in fname:
                fname = fname.split('@')[0]
            elif '$' in fname:
                fname = fname.split('$')[0]

            return (fname, num_basic_blocks, num_edges_of_blocks, num_call_subfunctions, func.body.numAddresses, func.parameterCount)

        def _syms_match(esym, esym2) -> Tuple[bool, str]:
            found = False
            match_type = None
            min_func_length = 15

            if esym2['name'] == esym['name'] and esym2['paramcount'] == esym['paramcount']:
                self.logger.info("Name + Paramcount {} {}".format(sym.getName(True), sym2.getName(True)))
                found = True
                match_type = 'Name:Param'
            if esym2['name'] == esym['name'] and esym2['length'] == esym['length']:
                self.logger.info("Name + length {} {}".format(sym.getName(True), sym2.getName(True)))
                found = True
                match_type = 'Name:Length'
            elif esym2['fullname'] == esym['fullname']:
                self.logger.info("Name Exact {} {}".format(sym.getName(True), sym2.getName(True)))
                found = True
                match_type = 'Fullname'

            return found, match_type

        from ghidra.program.model.symbol import SymbolType

        self.logger.info(f'Calculating initial structural signature for all ')
        p1_funcs = {}
        p2_funcs = {}

        from time import time
        start_p1 = time()
        for sym in p1.getSymbolTable().getDefinedSymbols():
            if sym.getSymbolType() == SymbolType.FUNCTION:
                func = p1.functionManager.getFunctionAt(sym.getAddress())
                p1_funcs[func] = _get_compare_key2(sym, func)
        end_p1 = time()

        for sym in p2.getSymbolTable().getDefinedSymbols():
            if sym.getSymbolType() == SymbolType.FUNCTION:
                func = p2.functionManager.getFunctionAt(sym.getAddress())
                p2_funcs[func] = _get_compare_key2(sym, func)
        end_p2 = time()

        self.logger.info(f'p1 structural signature time: {end_p1 - start_p1}')
        self.logger.info(f'p2 structural signature time: {end_p2 - end_p1}')

        from collections import Counter

        p1_funcs_vals = Counter(p1_funcs.values())
        p2_funcs_vals = Counter(p1_funcs.values())

        count_non_unique = 0
        for key, val in p1_funcs_vals.items():
            if val > 1:
                count_non_unique += 1
                self.logger.debug(f'{key, val}')
        self.logger.info(f'non-unique p1: {count_non_unique}')
        count_non_unique = 0
        for key, val in p2_funcs_vals.items():
            if val > 1:
                count_non_unique += 1
                self.logger.debug(f'{key, val}')
        self.logger.info(f'non-unique p2: {count_non_unique}')

        old_func_set = set(p1_funcs.values())
        new_func_set = set(p2_funcs.values())

        modified_old = sorted(old_func_set.difference(new_func_set))
        modified_new = sorted(new_func_set.difference(old_func_set))

        matching_compare_keys = sorted(old_func_set.intersection(new_func_set))
        self.logger.info(f'modified_old: {len(modified_old)}')
        self.logger.info(f'modified_new: {len(modified_new)}')
        self.logger.info(f'matching_compare_keys: {len(matching_compare_keys)}')

        p1_modified = []
        p2_modified = []

        p1_name_to_func = {}
        p2_name_to_func = {}

        for func, compare_key in p1_funcs.items():
            if compare_key in modified_old:
                p1_modified.append(func.getSymbol())
                key = (func.getName(True), func.parameterCount)
                p1_name_to_func.setdefault(key, []).append(func)

        for func, compare_key in p2_funcs.items():
            if compare_key in modified_new:
                p2_modified.append(func.getSymbol())
                key = (func.getName(True), func.parameterCount)
                p2_name_to_func.setdefault(key, []).append(func)

        self.logger.info("\nmodified_old_modified")
        for sym in p1_modified:
            self.logger.info(sym)

        self.logger.info("\nmodified_new_modified")
        for sym in p2_modified:
            self.logger.info(sym)

        # Find modified functions based on compare_key
        for sym in p1.getSymbolTable().getDefinedSymbols():

            if "function".lower() in sym.getSymbolType().toString().lower():
                func = p1.functionManager.getFunctionAt(sym.getAddress())
                if (_get_compare_key2(sym, func)) in modified_old:
                    p1_modified.append(sym)

        for sym in p2.getSymbolTable().getDefinedSymbols():

            if "function".lower() in sym.getSymbolType().toString().lower():
                func = p2.functionManager.getFunctionAt(sym.getAddress())
                if (_get_compare_key2(sym, func)) in modified_new:
                    p2_modified.append(sym)

        matched = []
        unmatched = []
        matches = []

        self.logger.info("\nMatching functions...")

        # match by name and paramcount
        for sym in p1_modified:

            if sym in matched:
                continue

            func = sym.getProgram().functionManager.getFunctionAt(sym.getAddress())

            key = (func.getName(True), func.parameterCount)

            if p2_name_to_func.get(key):

                func2 = p2_name_to_func[key][0]

                if len(p2_name_to_func[key]) > 1:

                    for func in p1_name_to_func[key]:

                        if func.getSymbol() in matched:
                            continue

                        for func2 in p2_name_to_func[key]:

                            if func2.getSymbol() in matched:
                                continue
                            # fname, num_basic_blocks, num_edges_of_blocks, num_call_subfunctions, func.body.numAddresses, func.parameterCount

                            _, _, _, func2_num_calls, _, _ = p2_funcs[func2]
                            _, _, _, func_num_calls, _, _ = p1_funcs[func]

                            if func2_num_calls == func_num_calls:
                                sym = func.getSymbol()
                                sym2 = func2.getSymbol()
                                self.logger.info(
                                    f"FullName + Paramcount + NumCalls {sym.getName(True)} {sym2.getName(True)}")
                                match_type = 'FullName:Param:NumCalls'
                                matched.append(sym)
                                matched.append(sym2)
                                matches.append([sym, sym2, match_type])

                else:
                    sym2 = func2.getSymbol()
                    self.logger.info(f"FullName + Paramcount {sym.getName(True)} {sym2.getName(True)}")
                    match_type = 'FullName:Param'
                    matched.append(sym)
                    matched.append(sym2)
                    matches.append([sym, sym2, match_type])

        for sym in p1_modified:
            found = False

            if sym in matched:
                continue

            sym2 = DiffUtility.getSymbol(sym, p2)

            if sym2 and sym.getName(True) == sym2.getName(True):
                found = True
                match_type = 'Direct'
                self.logger.info(f"direct getsymbol match {sym.getName(True)} {sym2.getName(True)}")
            else:
                for sym2 in p2_modified:
                    if sym2 in matched:
                        continue

                    func = sym.getProgram().functionManager.getFunctionAt(sym.getAddress())
                    ck1 = _get_compare_key(sym, func)

                    func2 = sym2.getProgram().functionManager.getFunctionAt(sym2.getAddress())
                    ck2 = _get_compare_key(sym2, func2)

                    if ck1[4] == ck2[4]:
                        found = True
                        match_type = 'MnemonicOrdered:Hash'
                        self.logger.info(f"{match_type} match {sym.getName(True)} {sym2.getName(True)}")

                    if found:
                        break

            if found:
                matched.append(sym)
                matched.append(sym2)
                matches.append([sym, sym2, match_type])
            else:
                self.logger.info(f"Deleted func found: {sym}")
                self.logger.info(
                    f"Deleted func found: {_get_compare_key(sym,sym.getProgram().functionManager.getFunctionAt(sym.getAddress()))}")
                unmatched.append(sym)
                # matched.append(sym)  # TODO check this

        for sym in p2_modified:
            found = False
            match_type = None

            if sym in matched:
                continue

            sym2 = DiffUtility.getSymbol(sym, p1)

            if sym2 and sym.getName(True) == sym2.getName(True):
                found = True
                match_type = 'Direct'
                self.logger.info(f"direct getsymbol match {sym.getName(True)} {sym2.getName(True)}")
            else:
                for sym2 in p1_modified:
                    if sym2 in matched:
                        continue

                    esym = self.enhance_sym(sym)
                    esym2 = self.enhance_sym(sym2)
                    found, match_type = _syms_match(esym, esym2)

                    if found:
                        break

            if found:
                matched.append(sym)
                matched.append(sym2)
                matches.append([sym, sym2, match_type])
            else:
                self.logger.info(f"Added func found: {sym}")
                unmatched.append(sym)

        self.logger.info(len(p1_modified))
        self.logger.info(len(p2_modified))

        matches = sorted(matches, key=lambda x: str(x[0]))

        # no need to skip types because we just ignore the matches for structual sig
        skip_types = []

        return [unmatched, matches, skip_types]



from collections import Counter
from time import time


from typing import List, Tuple, TYPE_CHECKING

from .ghidra_diff_engine import GhidraDiffEngine
from .implied_matches import correlate_implied_matches
from .decomp_correlate import decomp_correlate
from .bsim import correlate_bsim

if TYPE_CHECKING:
    import ghidra
    from ghidra_builtins import *


class VersionTrackingDiff(GhidraDiffEngine):
    """
    An Ghidra Diff implementation using several exact and some fuzzy correlators
    This differ is inspired by Ghidra's Version Tracking (albeit much faster)
    ghidra/tree/master/Ghidra/Features/VersionTracking
    """

    def find_matches(
        self,
        p1: "ghidra.program.model.listing.Program",
        p2: "ghidra.program.model.listing.Program",
    ) -> list:
        """
        Find matching and unmatched functions between p1 and p2
        """

        from ghidra.program.model.symbol import SymbolType
        from ghidra.program.model.address import AddressSet
        from ghidra.util.task import ConsoleTaskMonitor

        # Match functions
        from ghidra.app.plugin.match import MatchFunctions, MatchSymbol

        # Correlators
        from ghidra.app.plugin.match import ExactMnemonicsFunctionHasher, ExactBytesFunctionHasher, ExactInstructionsFunctionHasher
        from .correlators import StructuralGraphExactHasher, StructuralGraphHasher, BulkInstructionsHasher, BulkMnemonicHasher, BulkBasicBlockMnemonicHasher, NamespaceNameParamHasher, NameParamHasher, NameParamRefHasher, SigCallingCalledHasher, StringsRefsHasher, SwitchSigHasher, StrUniqueFuncRefsHasher

        monitor = ConsoleTaskMonitor()

        # loadedAndInitializedAddressSet
        #   /**
        #  * Returns the set of addresses which correspond to all the "loaded" memory blocks that have
        #  * initialized data.  This does not include initialized memory blocks that contain data from
        #  * the program's file header such as debug sections.
        #  */
        p1_unmatched = AddressSet(p1.memory.loadedAndInitializedAddressSet)
        p2_unmatched = AddressSet(p2.memory.loadedAndInitializedAddressSet)

        # keep track of all function mathces
        p1_matches = AddressSet()
        p2_matches = AddressSet()

        # tuples of correlators instances
        # ( name, hasher, one_to_one, one_to_many)
        # DO NOT CHANGE ORDER UNLESS INTENDED, ORDER HAS MAJOR IMPACT ON ACCURACY AND EFFICIENCY
        func_correlators = [
            ('ExactBytesFunctionHasher', ExactBytesFunctionHasher.INSTANCE, True, False),
            ('ExactInstructionsFunctionHasher', ExactInstructionsFunctionHasher.INSTANCE, True, False),
            (StructuralGraphExactHasher.MATCH_TYPE, StructuralGraphExactHasher(), True, False),
            ('ExactMnemonicsFunctionHasher', ExactMnemonicsFunctionHasher.INSTANCE, True, False),
            ('BSIM', None, True, False), # not a true function hasher
            (BulkInstructionsHasher.MATCH_TYPE, BulkInstructionsHasher(), True, False),
            (SigCallingCalledHasher.MATCH_TYPE, SigCallingCalledHasher(), True, False),
            (StringsRefsHasher.MATCH_TYPE, StringsRefsHasher(), True, False),
            (StrUniqueFuncRefsHasher.MATCH_TYPE, StrUniqueFuncRefsHasher(), True, False),
            (SwitchSigHasher.MATCH_TYPE, SwitchSigHasher(), True, False),
            # WARN: one_to_many=True flag allows for false negatives is structal graph matching. Mitgated by added references, func name in hash
            (StructuralGraphHasher.MATCH_TYPE, StructuralGraphHasher(), True, True),
            # WARN: one_to_many=True flag allows for false negatives
            (BulkBasicBlockMnemonicHasher.MATCH_TYPE, BulkBasicBlockMnemonicHasher(), True, True),
            (SigCallingCalledHasher.MATCH_TYPE, SigCallingCalledHasher(), True, False),
            (StringsRefsHasher.MATCH_TYPE, StringsRefsHasher(), True, False),
            (StrUniqueFuncRefsHasher.MATCH_TYPE, StrUniqueFuncRefsHasher(), True, False),
            (SwitchSigHasher.MATCH_TYPE, SwitchSigHasher(), True, False),
        ]

        unmatched = []
        matches = {}

        # Run Symbol Hash Correlator

        one_to_one = True
        include_externals = True
        min_sym_name_len = 3
        matchedSymbols = MatchSymbol.matchSymbol(
            p1, p1.getMemory(), p2, p2.getMemory(), min_sym_name_len, one_to_one, include_externals, monitor)

        start = time()
        name = 'SymbolsHash'
        skipped = 0
        for match in matchedSymbols:
            if match.matchType == SymbolType.FUNCTION:
                # sanity check symbolmatch
                func = p1.functionManager.getFunctionAt(match.aSymbolAddress)
                func2 = p2.functionManager.getFunctionAt(match.bSymbolAddress)
                if func.getName(True) != func2.getName(True):
                    skipped += 1
                    continue
                p1_matches.add(match.aSymbolAddress)
                p2_matches.add(match.bSymbolAddress)
                matches.setdefault((match.aSymbolAddress, match.bSymbolAddress), {}).setdefault(name, 0)
                matches[(match.aSymbolAddress, match.bSymbolAddress)][name] += 1
        end = time()

        p1_unmatched = p1_unmatched.subtract(p1_matches)
        p2_unmatched = p2_unmatched.subtract(p2_matches)

        self.logger.info(f'Exec time: {end-start:.4f} secs')
        self.logger.info(f'Match count {matchedSymbols.size() - skipped}')
        self.logger.info(Counter([tuple(x) for x in matches.values()]))


        # Run Function Hash Correlators
        # Each round of matching will "accept" the matches and subtract them from the unmatched functions
        # This is why the order of correlators matter
        for cor in func_correlators:            

            start = time()

            name, hasher, one_to_one, one_to_many = cor

            self.logger.info(f'Running correlator: {name}')
            self.logger.debug(f'hasher: {hasher}')
            self.logger.info(f'name: {name} one_to_one: {one_to_one} one_to_many: {one_to_many}')

            if name == 'BSIM':             
                if self.bsim_full:
                    # slower, but uses full adddress space for matching
                    correlate_bsim(matches, p1,p2, p1_matches, p2_matches, monitor, self.logger,enabled=self.bsim)
                else:
                    # only matches on functions that have no match
                    correlate_bsim(matches, p1,p2, p1_matches, p2_matches, monitor, self.logger, p1_addr_set=p1_unmatched, p2_addr_set=p2_unmatched, enabled=self.bsim)
            else:
                func_matches = MatchFunctions.matchFunctions(
                    p1, p1_unmatched, p2, p2_unmatched, self.min_func_len, one_to_one, one_to_many, hasher, monitor)

                for match in func_matches:
                    p1_matches.add(match.aFunctionAddress)
                    p2_matches.add(match.bFunctionAddress)
                    matches.setdefault((match.aFunctionAddress, match.bFunctionAddress), {}).setdefault(name, 0)
                    matches[(match.aFunctionAddress, match.bFunctionAddress)][name] += 1
                
                self.logger.info(f'Match count: {func_matches.size()}')

            end = time()

            # subtract unmatched functions
            p1_unmatched = p1_unmatched.subtract(p1_matches)
            p2_unmatched = p2_unmatched.subtract(p2_matches)

            self.logger.info(f'{name} Exec time: {end-start:.4f} secs')
            

            # kill noisy monitor after first run
            monitor = ConsoleTaskMonitor().DUMMY_MONITOR

        # Log current counts
        self.logger.info(Counter([tuple(x) for x in matches.values()]))

        monitor = ConsoleTaskMonitor()

        # Find unmatched functions

        p1_missing = self.get_funcs_from_addr_set(p1, p1_unmatched)
        p2_missing = self.get_funcs_from_addr_set(p2, p2_unmatched)

        # Find implied matches
        correlate_implied_matches(matches,
                                  p1_missing,
                                  p2_missing,
                                  p1_matches,
                                  p2_matches,
                                  p1,
                                  p2,
                                  self.max_workers,
                                  monitor,
                                  self.logger)

        p1_unmatched = p1_unmatched.subtract(p1_matches)
        p2_unmatched = p2_unmatched.subtract(p2_matches)

        p1_missing = self.get_funcs_from_addr_set(p1, p1_unmatched)
        p2_missing = self.get_funcs_from_addr_set(p2, p2_unmatched)

        # attempt to correlate amongst unmatched functions
        decomp_correlate(self, matches, p1_missing, p2_missing, p1_matches, p2_matches)

        p1_unmatched = p1_unmatched.subtract(p1_matches)
        p2_unmatched = p2_unmatched.subtract(p2_matches)

        p1_missing = self.get_funcs_from_addr_set(p1, p1_unmatched)
        p2_missing = self.get_funcs_from_addr_set(p2, p2_unmatched)

        unmatched.extend([func.getSymbol() for func in p1_missing])
        unmatched.extend([func.getSymbol() for func in p2_missing])

        self.logger.info(f'p1 missing = {len(p1_missing)}')
        self.logger.info(f'p2 missing = {len(p2_missing)}')

        # Debug umatched
        if self.logger.level == 10:  # debug level
            self.debug_function_hasher(unmatched, func_correlators, p1, p2)

        # Find external function unmatched and matched
        p1_externals = {}
        # get external funcs (these are still interesting)
        for func in p1.functionManager.getExternalFunctions():
            key = func.getName(True)
            # Apple Machos Specific (externals provide no relevant info)
            if "<EXTERNAL>::EXT_FUN_" in key:
                continue
            p1_externals[key] = func

        p2_externals = {}
        # get external funcs (these are still interesting)
        for func in p2.functionManager.getExternalFunctions():
            key = func.getName(True)
            # Apple Machos Specific (externals provide no relevant info)
            if "<EXTERNAL>::EXT_FUN_" in key:
                continue
            p2_externals[key] = func

        deleted_externs = list(set(p1_externals.keys()).difference(p2_externals.keys()))
        added_externs = list(set(p2_externals.keys()).difference(p1_externals.keys()))
        matched_externs = list(set(p2_externals.keys()).intersection(p1_externals.keys()))

        self.logger.info(f'Externs - deleted: {len(deleted_externs)} added: {len(added_externs)} matched:{len(matched_externs)}')

        unmatched.extend([p1_externals[key].getSymbol() for key in deleted_externs])
        unmatched.extend([p2_externals[key].getSymbol() for key in added_externs])

        for key in matched_externs:
            name = 'ExternalsName'
            func = p1_externals[key]
            func2 = p2_externals[key]
            matches.setdefault((func.entryPoint, func2.entryPoint), {}).setdefault(name, 0)
            matches[(func.entryPoint, func2.entryPoint)][name] += 1

        # translate matches to expected format [ sym, sym2, match_type ]
        matched = []
        for match_addrs, match_types in matches.items():

            func = p1.functionManager.getFunctionContaining(match_addrs[0])
            assert func.entryPoint == match_addrs[0]
            func2 = p2.functionManager.getFunctionContaining(match_addrs[1])
            assert func2.entryPoint == match_addrs[1]

            matched.append([func.getSymbol(), func2.getSymbol(), list(match_types.keys())])

        # skip types will undergo less processing
        skip_types = ['BulkBasicBlockMnemonicHash' 'Decomp Match']

        return [unmatched, matched, skip_types]

    def debug_function_hasher(self,
                              functions: list,
                              func_correlators: list,
                              p1: "ghidra.program.model.listing.Program",
                              p2: "ghidra.program.model.listing.Program"):
        """
        Debug unmatched functions throuh each function hasher
        """
        from ghidra.util.task import ConsoleTaskMonitor

        for cor in func_correlators:

            name, hasher, one_to_one, one_to_many = cor

            self.logger.debug(f'Debug umatched correlator: {name}')

            if hasattr(hasher, 'DEBUG'):
                hasher.DEBUG = True

            dummy = ConsoleTaskMonitor().DUMMY_MONITOR
            dummy.cancel()

            p1_hashes_seen = {}
            p2_hashes_seen = {}

            for sym in sorted(functions, key=lambda x: x.getProgram().getFunctionManager().getFunctionAt(x.address).body.numAddresses, reverse=True):
                func = sym.getProgram().getFunctionManager().getFunctionAt(sym.address)
                in_p1 = (func.getProgram().name == p1.name)

                if hasher is None:
                    continue

                func_hash = hasher.hash(func, dummy)
                func_len = func.getBody().getNumAddresses()

                if in_p1:
                    if func_hash in p1_hashes_seen.keys():
                        print(f'hash matches another func in p1? {func_hash} {name} {func_len}')
                    p1_hashes_seen.setdefault(func_hash, []).append(func)

                    if func_hash in p2_hashes_seen.keys():
                        print(f'why though? {func_hash} {name} {func_len}')
                        func_json = self.enhance_sym(func.getSymbol(), get_decomp_info=True)
                        p2_func_json = self.enhance_sym(p2_hashes_seen[func_hash][0].getSymbol(), get_decomp_info=True)
                        for key in func_json.keys():
                            print(func_json[key])
                            print(p2_func_json[key])
                else:
                    if func_hash in p2_hashes_seen:
                        print(f'hash matches another func in p2 {func_hash} {name} {func_len}')
                    p2_hashes_seen.setdefault(func_hash, []).append(func)

                    if func_hash in p1_hashes_seen:
                        print(f'why though? {func_hash} {name} {func_len}')

                self.logger.debug(f'{func.getProgram().getName()}:{func}:{func_hash}')



from pathlib import Path
import json
import pytest

from ghidriff import get_parser, get_engine_classes, VersionTrackingDiff, GhidraDiffEngine

SYMBOLS_DIR = 'symbols'
BINS_DIR = 'bins'


@pytest.mark.forked
def test_gdt_afd(shared_datadir: Path):
    """
    Tests application of a GDT to a program
    runs forked because each jpype jvm can only be initialized 1x
    """

    test_name = 'cve-2023-21768'
    output_path = shared_datadir / test_name
    output_path.mkdir(exist_ok=True, parents=True)
    symbols_path = shared_datadir / SYMBOLS_DIR
    bins_path = shared_datadir / BINS_DIR    
    ghidra_project_path = output_path / 'ghidra_projects'
    ghidra_project_path.mkdir(exist_ok=True,parents=True)
    gdt_path = (shared_datadir / 'ntddk_64.gdt')

    # setup bins
    old_bin_path = bins_path / 'afd.sys.x64.10.0.22621.1028'
    new_bin_path = bins_path / 'afd.sys.x64.10.0.22621.1415'

    assert old_bin_path.exists()
    assert new_bin_path.exists()

    parser = get_parser()

    GhidraDiffEngine.add_ghidra_args_to_parser(parser)

    args = parser.parse_args([
            '-s', 
            str(symbols_path), 
            str(old_bin_path.absolute()), 
            str(new_bin_path.absolute()),
            '-p', 
            str(ghidra_project_path.absolute()),
            "--gdt",
            str(gdt_path.absolute())
        ])        

    engine_log_path = output_path / parser.get_default('log_path')

    binary_paths = args.old + [bin for sublist in args.new for bin in sublist]

    binary_paths = [Path(path) for path in binary_paths]

    if any([not path.exists() for path in binary_paths]):
        missing_bins = [f'{path.name}' for path in binary_paths if not path.exists()]
        raise FileNotFoundError(f"Missing Bins: {' '.join(missing_bins)}")

    project_name = f'{args.project_name}-{binary_paths[0].name}-{binary_paths[-1].name}'

    DiffEngine: GhidraDiffEngine = VersionTrackingDiff

    d: GhidraDiffEngine = DiffEngine(args=args,
                                     verbose=True,
                                     threaded=args.threaded,
                                     max_ram_percent=args.max_ram_percent,
                                     print_jvm_flags=args.print_flags,
                                     jvm_args=args.jvm_args,
                                     force_analysis=args.force_analysis,
                                     force_diff=args.force_diff,
                                     verbose_analysis=args.va,
                                     no_symbols=args.no_symbols,
                                     engine_log_path=engine_log_path,
                                     engine_log_level=args.log_level,
                                     engine_file_log_level=args.file_log_level,
                                     gdts=args.gdt
                                     )

    d.setup_project(binary_paths, args.project_location, project_name, args.symbols_path)

    d.analyze_project()

    program = None
    for df in d.project.getRootFolder().getFiles():
        program = d.project.openProgram("/", df.getName(), False)
    
    ## without GDT this func sig return UNDEFINED and not types
    known_typed_sig = "BOOLEAN IoIs32bitProcess(PIRP Irp)"
    symbol_to_test = "IoIs32bitProcess"

    for f in program.functionManager.externalFunctions:

            if f'{f.getName()}' == symbol_to_test:
                print(f)
                signature_after_gdt = f'{f.getSignature()}'

    assert signature_after_gdt is not None
    assert signature_after_gdt == known_typed_sig


from ghidriff import GhidraDiffEngine
# from mock import patch
from pytest import MonkeyPatch
import pytest


def test_pyhidra_start():
    import pyhidra
    pyhidra.start(verbose=True)


def test_ghidra_install_dir():
    import os
    install_dir = os.getenv("GHIDRA_INSTALL_DIR")
    assert install_dir == "/ghidra"


# @pytest.fixture
# def setup_bogus_env(mon):


# @patch('pyhidra.GHIDRA_INSTALL_DIR', "/someboguspath"):
# def test_bogus_ghidra_install_dira():
#     with pytest.raises(FileNotFoundError):
#         import pyhidra as err_pyhidra
#         err_pyhidra.start(verbose=True)


# def test_ghidra_install_dir():
#     # import sys
#     # sys.modules.pop('pyhidra')
#     with MonkeyPatch.context() as mp:
#         mp.delenv("GHIDRA_INSTALL_DIR")

#         import pyhidra
#         # print(os.getenv("GHIDRA_INSTALL_DIR"))
#         with pytest.raises(SystemExit) as pytest_wrapped_e:
#             launcher = pyhidra.start(verbose=True)
#         assert pytest_wrapped_e.type == SystemExit
#         #assert pytest_wrapped_e.value.code == 42
#     import os
#     print(os.getenv("GHIDRA_INSTALL_DIR"))
#     # from importlib import reload
#     # reload(pyhidra)
#     import sys
#     del sys.modules['pyhidra']


# @pytest.mark.forked
# def test_bogus_ghidra_install_dir():
#     # monkeypatch.setenv("GHIDRA_INSTALL_DIR", '/someboguspath')
#     import os

#     with MonkeyPatch.context() as mp:
#         mp.delenv('GHIDRA_INSTALL_DIR')
#         print(os.getenv("GHIDRA_INSTALL_DIR"))

#         launcher = None

#         # with pytest.raises(SystemExit):
#         import pyhidra
#         mp.setattr(pyhidra.constants, 'GHIDRA_INSTALL_DIR', '/someboguspath')
#         print(pyhidra.constants)
#         print(os.getenv("GHIDRA_INSTALL_DIR"))
#         launcher = pyhidra.start(verbose=True)

#         assert launcher == None

    #     launcher = pyhidra.start(verbose=True)
    #     print(launcher.check_ghidra_version())

# det test_file_not_exist():



from pathlib import Path
import json
import pytest
from pyhidra import HeadlessPyhidraLauncher


from ghidriff import get_parser, get_engine_classes, VersionTrackingDiff, GhidraDiffEngine

SYMBOLS_DIR = 'symbols'
BINS_DIR = 'bins'


@pytest.mark.forked
def test_gzf_import_program(shared_datadir: Path):
    """
    Tests that binaries can be successfully imported
    Tests that gzf files contain expected programs
    """

    if HeadlessPyhidraLauncher().app_info.version < '11.0':
        # gzf files were made with 11.0
        print('Skip testing gzf on < 11.0')
        return

    test_name = 'test-imports'
    output_path = shared_datadir / test_name
    output_path.mkdir(exist_ok=True, parents=True)
    symbols_path = shared_datadir / SYMBOLS_DIR
    bins_path = shared_datadir / BINS_DIR
    ghidra_project_path = output_path / 'ghidra_projects'
    ghidra_project_path.mkdir(exist_ok=True,parents=True)

    # bins
    bins_to_import = [
        # bin path , expected program
        ['afd.sys.x64.10.0.22621.1028', 'afd.sys.x64.10.0.22621.1028-00a2b7'], #if a gzf file is used first, this becomes really unstable... 
        ['afd.sys.x64.10.0.22621.1415', 'afd.sys.x64.10.0.22621.1415-095200'],
        ['afd.sys.x64.10.0.22621.1028.gzf', 'afd.sys.x64.10.0.22621.1028.gzf-338a92'],        
        ['afd.sys.x64.10.0.22621.1415.gzf', 'afd.sys.x64.10.0.22621.1415.gzf-fc498a'],        
        ['ntoskrnl.exe.x64.10.0.22621.2792.10-1-5.gzf', 'ntoskrnl.exe.x64.10.0.22621.2792.10-1-5.gzf-acb020'],
        ['ntoskrnl.exe.x64.10.0.22621.2861.10-1-5.gzf', 'ntoskrnl.exe.x64.10.0.22621.2861.10-1-5.gzf-0e4e43'],
    ]

    parser = get_parser()

    GhidraDiffEngine.add_ghidra_args_to_parser(parser)

    engine_log_path = output_path / parser.get_default('log_path')
    
    binary_paths = [path for path in [bins_path / name[0] for name in bins_to_import ]]    
    
    args = parser.parse_args(['-s', str(symbols_path),'test', 'test2', '-p', str(ghidra_project_path.absolute())]) # these args will not be tested

    expected_names = [name for name in [name[1] for name in bins_to_import ]]
    
    binary_paths = [Path(path) for path in binary_paths]

    if any([not path.exists() for path in binary_paths]):
        missing_bins = [f'{path.name}' for path in binary_paths if not path.exists()]
        raise FileNotFoundError(f"Missing Bins: {' '.join(missing_bins)}")

    import uuid
    # ensure fresh test each time
    project_name = f'import-test-{uuid.uuid4()}'
    #project_name = f'import-test'

    DiffEngine: GhidraDiffEngine = VersionTrackingDiff

    d: GhidraDiffEngine = DiffEngine(args=args,
                                     verbose=True,
                                     threaded=args.threaded,
                                     max_ram_percent=args.max_ram_percent,
                                     print_jvm_flags=args.print_flags,
                                     jvm_args=args.jvm_args,
                                     force_analysis=args.force_analysis,
                                     force_diff=args.force_diff,
                                     verbose_analysis=args.va,
                                     no_symbols=args.no_symbols,
                                     engine_log_path=engine_log_path,
                                     engine_log_level=args.log_level,
                                     engine_file_log_level=args.file_log_level,
                                     )

    # imports_results = d.setup_project(binary_paths, args.project_location, project_name, args.symbols_path)

    # for i,data in enumerate(imports_results):
    #     # assert import names are as expected
    #     print(i)
    #     print(data)
    #     assert expected_names[i] == data[0]

    for i,import_path in enumerate(binary_paths):
        imports_result = d.setup_project([binary_paths[i]], args.project_location, project_name, args.symbols_path)
        #d.project.wait()
        assert expected_names[i] == imports_result[0][0]

